@@ -2,7 +2,8 @@
   +----------------------------------------------------------------------+
   | Suhosin Version 1                                                    |
   +----------------------------------------------------------------------+
-  | Copyright (c) 2006 The Hardened-PHP Project                          |
+  | Copyright (c) 2006-2007 The Hardened-PHP Project                     |
+  | Copyright (c) 2007 SektionEins GmbH                                  |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
@@ -12,11 +13,11 @@
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
-  | Author: Stefan Esser <sesser@hardened-php.net>                       |
+  | Author: Stefan Esser <sesser@sektioneins.de>                         |
   +----------------------------------------------------------------------+
 */
 
-/* $Id: execute.c,v 1.35 2006-11-14 14:39:48 sesser Exp $ */
+/* $Id: execute.c,v 1.1.1.1 2007-11-28 01:15:35 sesser Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -755,11 +756,173 @@ int ih_mail(IH_HANDLER_PARAMS)
 	return (0);
 }
 
+#define SQLSTATE_SQL        0
+#define SQLSTATE_IDENTIFIER 1
+#define SQLSTATE_STRING     2
+#define SQLSTATE_COMMENT    3
+#define SQLSTATE_MLCOMMENT  4
+
+int ih_querycheck(IH_HANDLER_PARAMS)
+{
+	void **p = EG(argument_stack).top_element-2;
+	unsigned long arg_count;
+	zval **arg;
+	char *query, *s, *e;
+	zval *backup;
+	int len;
+	char quote;
+	int state = SQLSTATE_SQL;
+	int cnt_union = 0, cnt_select = 0, cnt_comment = 0, cnt_opencomment = 0;
+	int mysql_extension = 0;
+
+	
+	SDEBUG("function: %s", ih->name);
+	arg_count = (unsigned long) *p;
+
+	if (ht < (long) ih->arg1) {
+		return (0);
+	}
+    
+	if ((long) ih->arg1) {
+    	    mysql_extension = 1;
+	}
+	
+	arg = (zval **) p - (arg_count - (long) ih->arg1 + 1); /* count from 0 */
+
+	backup = *arg;
+	if (Z_TYPE_P(backup) != IS_STRING) {
+		return (0);
+	}
+	len = Z_STRLEN_P(backup);
+	query = Z_STRVAL_P(backup);
+	
+	s = query;
+	e = s+len;
+	
+	while (s < e) {
+	    switch (state)
+	    {
+    		case SQLSTATE_SQL:
+    		    switch (s[0])
+    		    {
+        		case '`':
+        		    state = SQLSTATE_IDENTIFIER;
+        		    quote = '`';
+        		    break;
+        		case '\'':
+        		case '"':
+        		    state = SQLSTATE_STRING;
+        		    quote = *s;
+        		    break;
+        		case '/':
+        		    if (s[1]=='*') {
+                        if (mysql_extension == 1 && s[2] == '!') {
+                            s += 2;
+                            break;
+                        }
+            			s++;
+            			state = SQLSTATE_MLCOMMENT;
+        			    cnt_comment++;
+        		    }
+        		    break;
+    			case '-':
+        		    if (s[1]=='-') {
+        			s++;
+        			state = SQLSTATE_COMMENT;
+        			cnt_comment++;
+        		    }
+        		    break;
+    			case '#':
+        		    state = SQLSTATE_COMMENT;
+        		    cnt_comment++;
+        		    break;
+        		case 'u':
+    			case 'U':
+        		    if (strncasecmp("union", s, 5)==0) {
+            			s += 4;
+        			cnt_union++;
+        		    }
+        		    break;
+    			case 's':
+    			case 'S':
+        		    if (strncasecmp("select", s, 6)==0) {
+            			s += 5;
+        			cnt_select++;
+        		    }
+        		    break;
+    		    }
+    		    break;
+    		case SQLSTATE_STRING:
+		case SQLSTATE_IDENTIFIER:
+    		    if (s[0] == quote) {
+        		if (s[1] == quote) {
+        		    s++;
+    			} else {
+        		    state = SQLSTATE_SQL;
+    			}
+    		    }
+    		    if (s[0] == '\\') {
+    			s++;
+    		    }
+    		    break;
+		case SQLSTATE_COMMENT:
+    		    while (s[0] && s[0] != '\n') {
+    			s++;        
+    		    }
+    		    state = SQLSTATE_SQL;
+    		    break;
+    		case SQLSTATE_MLCOMMENT:
+    		    while (s[0] && (s[0] != '*' || s[1] != '/')) {
+    			s++;
+    		    }
+    		    if (s[0]) {
+    			state = SQLSTATE_SQL;
+    		    }
+    		    break;
+	    }
+	    s++;
+	}
+	if (state == SQLSTATE_MLCOMMENT) {
+	    cnt_opencomment = 1;
+	}
+	
+	if (cnt_opencomment && SUHOSIN_G(sql_opencomment)>0) {
+	    suhosin_log(S_SQL, "Open comment in SQL query: '%*s'", len, query);
+	    if (SUHOSIN_G(sql_opencomment)>1) {
+		suhosin_bailout(TSRMLS_C);
+	    }
+	}
+	
+	if (cnt_comment && SUHOSIN_G(sql_comment)>0) {
+	    suhosin_log(S_SQL, "Comment in SQL query: '%*s'", len, query);
+	    if (SUHOSIN_G(sql_comment)>1) {
+		suhosin_bailout(TSRMLS_C);
+	    }
+	}
+
+	if (cnt_union && SUHOSIN_G(sql_union)>0) {
+	    suhosin_log(S_SQL, "UNION in SQL query: '%*s'", len, query);
+	    if (SUHOSIN_G(sql_union)>1) {
+		suhosin_bailout(TSRMLS_C);
+	    }
+	}
+
+	if (cnt_select>1 && SUHOSIN_G(sql_mselect)>0) {
+	    suhosin_log(S_SQL, "Multiple SELECT in SQL query: '%*s'", len, query);
+	    if (SUHOSIN_G(sql_mselect)>1) {
+		suhosin_bailout(TSRMLS_C);
+	    }
+	}
+    
+	return (0);
+}
+
+
 int ih_fixusername(IH_HANDLER_PARAMS)
 {
 	void **p = EG(argument_stack).top_element-2;
 	unsigned long arg_count;
-	zval **arg;char *prefix, *postfix;
+	zval **arg;char *prefix, *postfix, *user;
 	zval *backup, *my_user;
 	int prefix_len, postfix_len, len;
 	
@@ -792,37 +955,146 @@ int ih_fixusername(IH_HANDLER_PARAMS)
 	arg = (zval **) p - (arg_count - (long) ih->arg1 + 1); /* count from 0 */
 
 	backup = *arg;
-	len = Z_STRLEN_P(backup);
+	if (Z_TYPE_P(backup) != IS_STRING) {
+		user = "";
+		len = 0;
+	} else {
+		len = Z_STRLEN_P(backup);
+		user = Z_STRVAL_P(backup);
+	}
 
 	if (prefix_len && prefix_len <= len) {
-		if (strncmp(prefix, Z_STRVAL_P(backup), prefix_len)==0) {
+		if (strncmp(prefix, user, prefix_len)==0) {
 			prefix = "";
 			len -= prefix_len;
 		}
 	}
 	
 	if (postfix_len && postfix_len <= len) {
-		if (strncmp(postfix, Z_STRVAL_P(backup)+Z_STRLEN_P(backup)-len, postfix_len)==0) {
+		if (strncmp(postfix, user+len-postfix_len, postfix_len)==0) {
 			postfix = "";
 		}
 	}
 	
 	MAKE_STD_ZVAL(my_user);
 	my_user->type = IS_STRING;
-	my_user->value.str.len = spprintf(&my_user->value.str.val, 0, "%s%s%s", prefix, Z_STRVAL_P(backup), postfix);
+	my_user->value.str.len = spprintf(&my_user->value.str.val, 0, "%s%s%s", prefix, user, postfix);
 	
 	/* XXX: memory_leak? */
 	*arg = my_user;	
 	 
-	SDEBUG("function: %s - user: %s", ih->name, Z_STRVAL_PP(arg));
+	SDEBUG("function: %s - user: %s", ih->name, user);
 
 	return (0);
 }
 
+static int suhosin_php_body_write(const char *str, uint str_length TSRMLS_DC)
+{
+#define P_META_ROBOTS "<meta name=\"ROBOTS\" content=\"NOINDEX,NOFOLLOW,NOARCHIVE\" />"
+#define S_META_ROBOTS "<meta name=\"ROBOTS\" content=\"NOINDEX,FOLLOW,NOARCHIVE\" />"
+
+    SDEBUG("bw: %s", str);
+
+	if ((str_length == sizeof("</head>\n")-1) && (strcmp(str, "</head>\n")==0)) {
+		SUHOSIN_G(old_php_body_write)(S_META_ROBOTS, sizeof(S_META_ROBOTS)-1 TSRMLS_CC);
+		OG(php_body_write) = SUHOSIN_G(old_php_body_write);
+		return SUHOSIN_G(old_php_body_write)(str, str_length TSRMLS_CC);
+	} else if ((str_length == sizeof(P_META_ROBOTS)-1) && (strcmp(str, P_META_ROBOTS)==0)) {
+		return str_length;
+	}
+	return SUHOSIN_G(old_php_body_write)(str, str_length TSRMLS_CC);	
+}
+
+static int ih_phpinfo(IH_HANDLER_PARAMS)
+{
+    int argc = ZEND_NUM_ARGS();
+	long flag;
+
+	if (zend_parse_parameters(argc TSRMLS_CC, "|l", &flag) == FAILURE) {
+        RETVAL_FALSE;
+		return (1);
+	}
+
+	if(!argc) {
+		flag = PHP_INFO_ALL;
+	}
+
+	/* Andale!  Andale!  Yee-Hah! */
+	php_start_ob_buffer(NULL, 4096, 0 TSRMLS_CC);
+	if (!sapi_module.phpinfo_as_text) {
+		SUHOSIN_G(old_php_body_write) = OG(php_body_write);
+		OG(php_body_write) = suhosin_php_body_write;
+	}
+	php_print_info(flag TSRMLS_CC);
+	php_end_ob_buffer(1, 0 TSRMLS_CC);
+
+	RETVAL_TRUE;
+	return (1);
+}
+
+
+static int ih_function_exists(IH_HANDLER_PARAMS)
+{
+	zval **function_name;
+	zend_function *func;
+	char *lcname;
+	zend_bool retval;
+	int func_name_len;
+	
+	if (ZEND_NUM_ARGS()!=1 || zend_get_parameters_ex(1, &function_name)==FAILURE) {
+		ZEND_WRONG_PARAM_COUNT();
+	}
+	convert_to_string_ex(function_name);
+	func_name_len = Z_STRLEN_PP(function_name);
+	lcname = estrndup(Z_STRVAL_PP(function_name), func_name_len);
+	zend_str_tolower(lcname, func_name_len);
+
+	retval = (zend_hash_find(EG(function_table), lcname, func_name_len+1, (void **)&func) == SUCCESS);
+	
+	efree(lcname);
+
+	/*
+	 * A bit of a hack, but not a bad one: we see if the handler of the function
+	 * is actually one that displays "function is disabled" message.
+	 */
+	if (retval && func->type == ZEND_INTERNAL_FUNCTION &&
+		func->internal_function.handler == zif_display_disabled_function) {
+		retval = 0;
+	}
+
+	/* Now check if function is forbidden by Suhosin */
+	if (SUHOSIN_G(in_code_type) == SUHOSIN_EVAL) {
+		if (SUHOSIN_G(eval_whitelist) != NULL) {
+			if (!zend_hash_exists(SUHOSIN_G(eval_whitelist), lcname, func_name_len+1)) {
+			    retval = 0;
+			}
+		} else if (SUHOSIN_G(eval_blacklist) != NULL) {
+			if (zend_hash_exists(SUHOSIN_G(eval_blacklist), lcname, func_name_len+1)) {
+			    retval = 0;
+			}
+		}
+	}
+	
+	if (SUHOSIN_G(func_whitelist) != NULL) {
+		if (!zend_hash_exists(SUHOSIN_G(func_whitelist), lcname, func_name_len+1)) {
+		    retval = 0;
+		}
+	} else if (SUHOSIN_G(func_blacklist) != NULL) {
+		if (zend_hash_exists(SUHOSIN_G(func_blacklist), lcname, func_name_len+1)) {
+		    retval = 0;
+		}
+	}
+
+	RETVAL_BOOL(retval);
+	return (1);
+}
+
+
 internal_function_handler ihandlers[] = {
     { "preg_replace", ih_preg_replace, NULL, NULL, NULL },
     { "mail", ih_mail, NULL, NULL, NULL },
     { "symlink", ih_symlink, NULL, NULL, NULL },
+    { "phpinfo", ih_phpinfo, NULL, NULL, NULL },
 	
     { "ocilogon", ih_fixusername, (void *)1, NULL, NULL },
     { "ociplogon", ih_fixusername, (void *)1, NULL, NULL },
@@ -834,6 +1106,8 @@ internal_function_handler ihandlers[] = {
     { "fbsql_change_user", ih_fixusername, (void *)1, NULL, NULL },
     { "fbsql_connect", ih_fixusername, (void *)2, NULL, NULL },
     { "fbsql_pconnect", ih_fixusername, (void *)2, NULL, NULL },
+    
+    { "function_exists", ih_function_exists, NULL, NULL, NULL },
 	
     { "ifx_connect", ih_fixusername, (void *)2, NULL, NULL },
     { "ifx_pconnect", ih_fixusername, (void *)2, NULL, NULL },
@@ -850,6 +1124,15 @@ internal_function_handler ihandlers[] = {
     { "mssql_connect", ih_fixusername, (void *)2, NULL, NULL },
     { "mssql_pconnect", ih_fixusername, (void *)2, NULL, NULL },
 
+    { "mysql_query", ih_querycheck, (void *)1, (void *)1, NULL },
+    { "mysql_db_query", ih_querycheck, (void *)2, (void *)1, NULL },
+    { "mysql_unbuffered_query", ih_querycheck, (void *)1, (void *)1, NULL },
+    { "mysqli_query", ih_querycheck, (void *)2, (void *)1, NULL },
+    { "mysqli_real_query", ih_querycheck, (void *)2, (void *)1, NULL },
+    { "mysqli_send_query", ih_querycheck, (void *)2, (void *)1, NULL },
+    { "mysqli_master_query", ih_querycheck, (void *)2, (void *)1, NULL },
+    { "mysqli_slave_query", ih_querycheck, (void *)2, (void *)1, NULL },
+
     { "mysqli", ih_fixusername, (void *)2, NULL, NULL },
     { "mysql_connect", ih_fixusername, (void *)2, NULL, NULL },
     { "mysql_pconnect", ih_fixusername, (void *)2, NULL, NULL },
@@ -858,20 +1141,41 @@ internal_function_handler ihandlers[] = {
     { NULL, NULL, NULL, NULL, NULL }
 };
 
+#define FUNCTION_WARNING() zend_error(E_WARNING, "%s() has been disabled for security reasons", get_active_function_name(TSRMLS_C));
 
 /* {{{ void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int return_value_used TSRMLS_DC)
  *    This function provides a hook for internal execution */
 static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int return_value_used TSRMLS_DC)
 {
 	char *lcname;
-	int function_name_strlen;
+	int function_name_strlen, free_lcname = 0;
 	zval *return_value;
+	zend_class_entry *ce;
 	int ht;
 	internal_function_handler *ih;
 	
+	ce = ((zend_internal_function *) execute_data_ptr->function_state.function)->scope;
 	lcname = ((zend_internal_function *) execute_data_ptr->function_state.function)->function_name;
 	function_name_strlen = strlen(lcname);
+	
+	/* handle methodcalls correctly */
+	if (ce != NULL) {
+		char *tmp = (char *) emalloc(function_name_strlen + 2 + ce->name_length + 1);
+		memcpy(tmp, ce->name, ce->name_length);
+		memcpy(tmp+ce->name_length, "::", 2);
+		memcpy(tmp+ce->name_length+2, lcname, function_name_strlen);
+		lcname = tmp;
+		free_lcname = 1;
+		function_name_strlen += ce->name_length + 2;
+		lcname[function_name_strlen] = 0;
+		zend_str_tolower(lcname, function_name_strlen);
+	}
+	
+#ifdef ZEND_ENGINE_2  
 	return_value = (*(temp_variable *)((char *) execute_data_ptr->Ts + execute_data_ptr->opline->result.u.var)).var.ptr;
+#else
+    return_value = execute_data_ptr->Ts[execute_data_ptr->opline->result.u.var].var.ptr;
+#endif
 	ht = execute_data_ptr->opline->extended_value;
 
 	SDEBUG("function: %s", lcname);
@@ -880,32 +1184,37 @@ static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int re
 	
 		if (SUHOSIN_G(eval_whitelist) != NULL) {
 			if (!zend_hash_exists(SUHOSIN_G(eval_whitelist), lcname, function_name_strlen+1)) {
-			    suhosin_log(S_EXECUTOR, "function outside of eval whitelist called: %s()", lcname);
-			    suhosin_bailout(TSRMLS_C);
+				suhosin_log(S_EXECUTOR, "function outside of eval whitelist called: %s()", lcname);
+				goto execute_internal_bailout;
 			}
 		} else if (SUHOSIN_G(eval_blacklist) != NULL) {
 			if (zend_hash_exists(SUHOSIN_G(eval_blacklist), lcname, function_name_strlen+1)) {
-			    suhosin_log(S_EXECUTOR, "function within eval blacklist called: %s()", lcname);
-			    suhosin_bailout(TSRMLS_C);
+				suhosin_log(S_EXECUTOR, "function within eval blacklist called: %s()", lcname);
+				goto execute_internal_bailout;
 			}
 		}
 	}
 	
 	if (SUHOSIN_G(func_whitelist) != NULL) {
 		if (!zend_hash_exists(SUHOSIN_G(func_whitelist), lcname, function_name_strlen+1)) {
-		    suhosin_log(S_EXECUTOR, "function outside of whitelist called: %s()", lcname);
-		    suhosin_bailout(TSRMLS_C);
+			suhosin_log(S_EXECUTOR, "function outside of whitelist called: %s()", lcname);
+			goto execute_internal_bailout;
 		}
 	} else if (SUHOSIN_G(func_blacklist) != NULL) {
 		if (zend_hash_exists(SUHOSIN_G(func_blacklist), lcname, function_name_strlen+1)) {
-		    suhosin_log(S_EXECUTOR, "function within blacklist called: %s()", lcname);
-		    suhosin_bailout(TSRMLS_C);
+			suhosin_log(S_EXECUTOR, "function within blacklist called: %s()", lcname);
+			goto execute_internal_bailout;
 		}
 	}
 	
 	if (zend_hash_find(&ihandler_table, lcname, function_name_strlen+1, (void **)&ih) == SUCCESS) {
 	
-		int retval = ih->handler(IH_HANDLER_PARAM_PASSTHRU);
+		int retval = 0;
+		void *handler = ((zend_internal_function *) execute_data_ptr->function_state.function)->handler;
+		
+		if (handler != ZEND_FN(display_disabled_function)) {
+		    retval = ih->handler(IH_HANDLER_PARAM_PASSTHRU);
+		}
 		
 		if (retval == 0) {
 			old_execute_internal(execute_data_ptr, return_value_used TSRMLS_CC);
@@ -913,6 +1222,16 @@ static void suhosin_execute_internal(zend_execute_data *execute_data_ptr, int re
 	} else {
 		old_execute_internal(execute_data_ptr, return_value_used TSRMLS_CC);
 	}
+	if (free_lcname == 1) {
+		efree(lcname);
+	}
+	return;
+execute_internal_bailout:
+	if (free_lcname == 1) {
+		efree(lcname);
+	}
+	FUNCTION_WARNING()
+	suhosin_bailout(TSRMLS_C);
 }
 /* }}} */
 
